@@ -12,6 +12,7 @@ if hasattr(rmm, "DeviceBuffer"):
     @cuda_serialize.register(rmm.DeviceBuffer)
     def cuda_serialize_rmm_device_buffer(x):
         header = x.__cuda_array_interface__.copy()
+        header["strides"] = (1,)
         header["lengths"] = [x.nbytes]
         frames = [x]
         return header, frames
@@ -30,6 +31,7 @@ if hasattr(rmm, "DeviceBuffer"):
     @dask_serialize.register(rmm.DeviceBuffer)
     def dask_serialize_rmm_device_buffer(x):
         header, frames = cuda_serialize_rmm_device_buffer(x)
+        header["writeable"] = (None,) * len(frames)
         frames = [numba.cuda.as_cuda_array(f).copy_to_host().data for f in frames]
         return header, frames
 
