@@ -1,15 +1,17 @@
-import uuid
-from collections import defaultdict, deque
 import asyncio
-import dask
+import logging
+import uuid
+import warnings
 from asyncio import TimeoutError
+from collections import defaultdict, deque
+
+import dask
 from tornado.ioloop import PeriodicCallback
+
+from distributed.utils_comm import retry_operation
+from .metrics import time
 from .utils import log_errors, parse_timedelta
 from .worker import get_client
-from .metrics import time
-import warnings
-import logging
-from distributed.utils_comm import retry_operation
 
 logger = logging.getLogger(__name__)
 
@@ -33,15 +35,15 @@ class _Watch:
 
 
 class SemaphoreExtension:
-    """ An extension for the scheduler to manage Semaphores
+    """An extension for the scheduler to manage Semaphores
 
     This adds the following routes to the scheduler
 
     * semaphore_acquire
     * semaphore_release
-    * semaphore_create
     * semaphore_close
     * semaphore_refresh_leases
+    * semaphore_register
     """
 
     def __init__(self, scheduler):
@@ -56,7 +58,7 @@ class SemaphoreExtension:
 
         self.scheduler.handlers.update(
             {
-                "semaphore_create": self.create,
+                "semaphore_register": self.create,
                 "semaphore_acquire": self.acquire,
                 "semaphore_release": self.release,
                 "semaphore_close": self.close,
@@ -67,16 +69,24 @@ class SemaphoreExtension:
 
         self.scheduler.extensions["semaphores"] = self
 
+        # {metric_name: {semaphore_name: metric}}
+        self.metrics = {
+            "acquire_total": defaultdict(int),  # counter
+            "release_total": defaultdict(int),  # counter
+            "average_pending_lease_time": defaultdict(float),  # gauge
+            "pending": defaultdict(int),  # gauge
+        }
+
         validation_callback_time = parse_timedelta(
             dask.config.get("distributed.scheduler.locks.lease-validation-interval"),
             default="s",
         )
         self._pc_lease_timeout = PeriodicCallback(
-            self._check_lease_timeout, validation_callback_time * 1000,
+            self._check_lease_timeout, validation_callback_time * 1000
         )
         self._pc_lease_timeout.start()
         self.lease_timeout = parse_timedelta(
-            dask.config.get("distributed.scheduler.locks.lease-timeout"), default="s",
+            dask.config.get("distributed.scheduler.locks.lease-timeout"), default="s"
         )
 
     async def get_value(self, comm=None, name=None):
@@ -122,6 +132,7 @@ class SemaphoreExtension:
             now = time()
             logger.info("Acquire lease %s for %s at %s", lease_id, name, now)
             self.leases[name][lease_id] = now
+            self.metrics["acquire_total"][name] += 1
         else:
             result = False
         return result
@@ -141,6 +152,7 @@ class SemaphoreExtension:
             w = _Watch(timeout)
             w.start()
 
+            self.metrics["pending"][name] += 1
             while True:
                 logger.info(
                     "Trying to acquire %s for %s with %ss left.",
@@ -172,6 +184,12 @@ class SemaphoreExtension:
                     result,
                     w.elapsed(),
                 )
+                # We're about to return, so the lease is no longer "pending"
+                self.metrics["average_pending_lease_time"][name] = (
+                    self.metrics["average_pending_lease_time"][name] + w.elapsed()
+                ) / 2
+                self.metrics["pending"][name] -= 1
+
                 return result
 
     def release(self, comm=None, name=None, lease_id=None):
@@ -196,6 +214,7 @@ class SemaphoreExtension:
         # Everything needs to be atomic here.
         del self.leases[name][lease_id]
         self.events[name].set()
+        self.metrics["release_total"][name] += 1
 
     def _check_lease_timeout(self):
         now = time()
@@ -235,10 +254,20 @@ class SemaphoreExtension:
                         RuntimeWarning,
                     )
                 del self.leases[name]
+            if name in self.metrics["pending"]:
+                if self.metrics["pending"][name]:
+                    warnings.warn(
+                        f"Closing semaphore {name} but there remain pending leases",
+                        RuntimeWarning,
+                    )
+            # Clean-up state of semaphore metrics
+            for _, metric_dict in self.metrics.items():
+                if name in metric_dict:
+                    del metric_dict[name]
 
 
 class Semaphore:
-    """ Semaphore
+    """Semaphore
 
     This `semaphore <https://en.wikipedia.org/wiki/Semaphore_(programming)>`_
     will track leases on the scheduler which can be acquired and
@@ -285,22 +314,28 @@ class Semaphore:
     client: Client (optional)
         Client to use for communication with the scheduler.  If not given, the
         default global client will be used.
+    register: bool
+        If True, register the semaphore with the scheduler. This needs to be
+        done before any leases can be acquired. If not done during
+        initialization, this can also be done by calling the register method of
+        this class.
+        When registering, this needs to be awaited.
 
     Examples
     --------
     >>> from distributed import Semaphore
-    >>> sem = Semaphore(max_leases=2, name='my_database')
-    >>>
-    >>> def access_resource(s, sem):
-    >>>     # This automatically acquires a lease from the semaphore (if available) which will be
-    >>>     # released when leaving the context manager.
-    >>>     with sem:
-    >>>         pass
-    >>>
-    >>> futures = client.map(access_resource, range(10), sem=sem)
-    >>> client.gather(futures)
-    >>> # Once done, close the semaphore to clean up the state on scheduler side.
-    >>> sem.close()
+    ... sem = Semaphore(max_leases=2, name='my_database')
+    ...
+    ... def access_resource(s, sem):
+    ...     # This automatically acquires a lease from the semaphore (if available) which will be
+    ...     # released when leaving the context manager.
+    ...     with sem:
+    ...         pass
+    ...
+    ... futures = client.map(access_resource, range(10), sem=sem)
+    ... client.gather(futures)
+    ... # Once done, close the semaphore to clean up the state on scheduler side.
+    ... sem.close()
 
     Notes
     -----
@@ -320,18 +355,19 @@ class Semaphore:
 
     """
 
-    def __init__(self, max_leases=1, name=None, client=None):
+    def __init__(self, max_leases=1, name=None, client=None, register=True):
         self.client = client or get_client()
         self.name = name or "semaphore-" + uuid.uuid4().hex
         self.max_leases = max_leases
         self.id = uuid.uuid4().hex
         self._leases = deque()
 
-        self._started = self.client.sync(
-            self.client.scheduler.semaphore_create,
-            name=self.name,
-            max_leases=max_leases,
-        )
+        self.refresh_leases = True
+
+        self._registered = None
+        if register:
+            self._registered = self.register()
+
         # this should give ample time to refresh without introducing another
         # config parameter since this *must* be smaller than the timeout anyhow
         refresh_leases_interval = (
@@ -341,7 +377,6 @@ class Semaphore:
             )
             / 5
         )
-        self._refreshing_leases = False
         pc = PeriodicCallback(
             self._refresh_leases, callback_time=refresh_leases_interval * 1000
         )
@@ -349,12 +384,29 @@ class Semaphore:
         # Registering the pc to the client here is important for proper cleanup
         self._periodic_callback_name = f"refresh_semaphores_{self.id}"
         self.client._periodic_callbacks[self._periodic_callback_name] = pc
-        pc.start()
-        self.refresh_leases = True
+
+        # Need to start the callback using IOLoop.add_callback to ensure that the
+        # PC uses the correct event loop.
+        self.client.io_loop.add_callback(pc.start)
+
+    def register(self):
+        """
+        Register the semaphore on scheduler side
+
+        This will register the semaphore on scheduler side and ensure that all necessary data structures exist.
+        """
+        if self._registered is None:
+            self._registered = self.client.sync(
+                self.client.scheduler.semaphore_register,
+                name=self.name,
+                max_leases=self.max_leases,
+            )
+        return self._registered
 
     def __await__(self):
         async def create_semaphore():
-            await self._started
+            if self._registered:
+                await self._registered
             return self
 
         return create_semaphore().__await__()
@@ -374,7 +426,7 @@ class Semaphore:
     async def _acquire(self, timeout=None):
         lease_id = uuid.uuid4().hex
         logger.info(
-            "%s requests lease for %s with ID %s", self.client.id, self.name, lease_id,
+            "%s requests lease for %s with ID %s", self.client.id, self.name, lease_id
         )
 
         # Using a unique lease id generated here allows us to retry since the
@@ -385,6 +437,8 @@ class Semaphore:
             name=self.name,
             timeout=timeout,
             lease_id=lease_id,
+            operation="semaphore acquire: client=%s, lease_id=%s, name=%s"
+            % (self.client.id, lease_id, self.name),
         )
         if result:
             self._leases.append(lease_id)
@@ -396,25 +450,56 @@ class Semaphore:
 
         If the internal counter is greater than zero, decrement it by one and return True immediately.
         If it is zero, wait until a release() is called and return True.
+
+        Parameters
+        ----------
+        timeout : number or string or timedelta, optional
+            Seconds to wait on acquiring the semaphore.  This does not
+            include local coroutine time, network transfer time, etc..
+            Instead of number of seconds, it is also possible to specify
+            a timedelta in string format, e.g. "200ms".
         """
+        timeout = parse_timedelta(timeout)
         return self.client.sync(self._acquire, timeout=timeout)
 
-    def release(self):
-        """
-        Release a semaphore.
-
-        Increment the internal counter by one.
-        """
-
-        """ Release the lock if already acquired """
-        if not self._leases:
-            raise RuntimeError("Released too often")
+    async def _release(self):
         # popleft to release the oldest lease first
         lease_id = self._leases.popleft()
         logger.info("%s releases %s for %s", self.client.id, lease_id, self.name)
-        return self.client.sync(
-            self.client.scheduler.semaphore_release, name=self.name, lease_id=lease_id,
-        )
+
+        try:
+            await retry_operation(
+                self.client.scheduler.semaphore_release,
+                name=self.name,
+                lease_id=lease_id,
+                operation="semaphore release: client=%s, lease_id=%s, name=%s"
+                % (self.client.id, lease_id, self.name),
+            )
+            return True
+        except Exception:  # Release fails for whatever reason
+            logger.error(
+                "Release failed for client=%s, lease_id=%s, name=%s. Cluster network might be unstable?"
+                % (self.client.id, lease_id, self.name),
+                exc_info=True,
+            )
+            return False
+
+    def release(self):
+        """
+        Release the semaphore.
+
+        Returns
+        -------
+        bool
+            This value indicates whether a lease was released immediately or not. Note that a user should  *not* retry
+            this operation. Under certain circumstances (e.g. scheduler overload) the lease may not be released
+            immediately, but it will always be automatically released after a specific interval configured using
+            "distributed.scheduler.locks.lease-validation-interval" and "distributed.scheduler.locks.lease-timeout".
+        """
+        if not self._leases:
+            raise RuntimeError("Released too often")
+
+        return self.client.sync(self._release)
 
     def get_value(self):
         """
@@ -444,7 +529,7 @@ class Semaphore:
     def __setstate__(self, state):
         name, max_leases = state
         client = get_client()
-        self.__init__(name=name, client=client, max_leases=max_leases)
+        self.__init__(name=name, client=client, max_leases=max_leases, register=False)
 
     def close(self):
         return self.client.sync(self.client.scheduler.semaphore_close, name=self.name)
